/* Version 0.23 2007.08.06 */
/* Local RCPT checking */
/* Author: Sergey S. Dukachyov, <admin@mfitness.ru> */
/* License: GPL v3 */


#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <signal.h>
#include <sysexits.h>
#include <unistd.h>
#include "libpq-fe.h"

#include "libmilter/mfapi.h"

#ifndef true
typedef int bool;
# define false	0
# define true	1
#endif /* ! true */

extern sfsistat	 mlfi_cleanup(SMFICTX *, bool);

//     
typedef struct 
  {
    PGconn *conn;
    char *clean_address;
    char *clean_domain_name;
    char *temp;
  }  cortege;

bool debug;
char *conninfo;
bool ignore_mail_from_cheking = false;
bool ignore_rcpt_to_cheking = false;


static bool address_preparation(register char *dst, register const char *src) {
    register const char *start = NULL, *stop = NULL;
    int tail;

    if (!(start = strchr(src, '<'))) return false;
    if (!(stop = strrchr(src, '>'))) return false;
    if (++start >= --stop) return false;
    strncpy(dst, start, stop - start + 1);
    tail = strlen(dst) - 1;
    if ((dst[0] >= 0x07 && dst[0] <= 0x0d) || dst[0] == 0x20) return false;
    if ((dst[tail] >= 0x07 && dst[tail] <= 0x0d) || dst[tail] == 0x20) return false;
    if (!strchr(dst, '@')) return false;
    return true;
}

static bool domain_extraction(register char *dst, register const char *src) {
    register const char *start = NULL;
    if (!(start = strchr(src, '@'))) return false;
    strcpy(dst, ++start);
    return true;
}


void disconnect_from_db(SMFICTX *ctx)
  {
  cortege *var = (cortege *)smfi_getpriv(ctx);
  if(var && var->conn) 
    {
    PQfinish(var->conn);
    var->conn = NULL;
    if(debug) syslog(LOG_NOTICE, "Connection to database closed.");
    }
  return;
  }

void free_cortege_memory(SMFICTX *ctx)
  {
  cortege *var = (cortege *)smfi_getpriv(ctx);
  if(!var) return;
  if(var->temp) {free(var->temp); var->temp = NULL;
        if(debug) syslog(LOG_NOTICE, "Free \'temp\'.");}
  if(var->clean_domain_name) {free(var->clean_domain_name); var->clean_domain_name = NULL;
        if(debug) syslog(LOG_NOTICE, "Free \'clean_domain_name\'.");}
  if(var->clean_address) {free(var->clean_address); var->clean_address = NULL;
        if(debug) syslog(LOG_NOTICE, "Free \'clean_address\'.");}
  return;
  }


sfsistat
check_local_user_existing(ctx, envto, typeofaddress, bounce)
	SMFICTX *ctx;
	char **envto;
	char* typeofaddress;
	char *bounce;
{
    PGresult   *res;

    cortege *variables = (cortege *)smfi_getpriv(ctx);
    if (!variables)
      {
      variables = malloc(sizeof(cortege));
      if(!variables)
        {
        if(debug) syslog(LOG_NOTICE, "Can't allocate memory for cortege.");
        smfi_setreply(ctx, "550", "4.3.2", "Internal system error");
	return SMFIS_TEMPFAIL;
	}
      memset(variables, 0, sizeof(cortege));
      smfi_setpriv(ctx, (void *) variables);     
      }
      
//  
        free_cortege_memory(ctx);

	variables->clean_address = malloc(strlen(envto[0]) + 1);
	if (!variables->clean_address)
	{
		/* can't accept this message right now */
	return SMFIS_TEMPFAIL;
	}
	memset(variables->clean_address, 0, strlen(envto[0]) + 1);

	variables->clean_domain_name = malloc(strlen(envto[0]) + 1);
	if (!variables->clean_domain_name)
	{
	/* can't accept this message right now */
	  return SMFIS_TEMPFAIL;
	}
	memset(variables->clean_domain_name, 0, strlen(envto[0]) + 1);

    if(debug) syslog(LOG_NOTICE, "Memory allocated.");

    if (!address_preparation(variables->clean_address, envto[0])) {
            smfi_setreply(ctx, "550", "5.1.3", "E-mail address does not conform to RFC-2821 syntax");
            return SMFIS_REJECT;
        }
//    !
//    strlwr(clean_address);

    if(debug) syslog(LOG_NOTICE, "%s: <%s>", typeofaddress, variables->clean_address);

    if (!domain_extraction(variables->clean_domain_name, variables->clean_address)) {
            smfi_setreply(ctx, "550", "5.1.1", "Trouble with domain extraction from e-mail address");
            return SMFIS_REJECT;
        }

if (!variables->conn)
  {
  variables->conn = PQconnectdb((char *)conninfo);

     if (PQstatus(variables->conn) != CONNECTION_OK)
       {
       if(debug) syslog(LOG_NOTICE, "Connection to database failed: %s", PQerrorMessage(variables->conn));
       smfi_setreply(ctx, "550", "4.3.2", "Can't check e-mail address.");
       return SMFIS_TEMPFAIL;
       }
  if(debug) syslog(LOG_NOTICE, "Connect to database.");
  }


    variables->temp = malloc (320 + strlen(variables->clean_address) + strlen(bounce));
    if(!variables->temp)
      {
      if(debug) syslog(LOG_NOTICE, "Cant't allocate memory to \'temp\' variable.");
      return SMFIS_TEMPFAIL;
      }
//   
   sprintf(variables->temp, "SELECT s_in FROM mailertable WHERE s_in=lower(\'%s\')", variables->clean_domain_name);
   if(debug) syslog(LOG_NOTICE, "SELECT command: %s", variables->temp);
   res = PQexecParams(variables->conn,
    variables->temp,
    0,       /* no param */
    NULL,    /* let the backend deduce param type */
    NULL,
    NULL,    /* don't need param lengths since text */
    NULL,    /* default to all text params */
    1);      /* ask for binary results */


  if (PQresultStatus(res) != PGRES_TUPLES_OK)
    {
    if(debug) syslog(LOG_NOTICE, "SELECT failed: %s", PQerrorMessage(variables->conn));
    fprintf(stderr, "SELECT failed: %s", PQerrorMessage(variables->conn));
    PQclear(res);
    return SMFIS_TEMPFAIL;
    }
  int PQntuples_var = PQntuples(res);
  PQclear(res);

  if(debug) syslog(LOG_NOTICE, "SELECT Success. PQntuples(res)=%d", PQntuples_var);
  
    if (PQntuples_var == 0) {
//     -        
      if(debug) syslog(LOG_NOTICE, "%s domain: <%s>. Alien. Continue checking on next stage", typeofaddress, variables->clean_domain_name);
      return SMFIS_CONTINUE;
      }

    if(debug) syslog(LOG_NOTICE, "%s domain: <%s>. Local.", typeofaddress, variables->clean_domain_name);
      

//       
    sprintf(variables->temp, "SELECT account FROM users WHERE account=lower(\'%s\')", variables->clean_address);
   if(debug) syslog(LOG_NOTICE, "SELECT command: %s", variables->temp);
  res = PQexecParams(variables->conn,
    variables->temp,
    0,       /* no param */
    NULL,    /* let the backend deduce param type */
    NULL,
    NULL,    /* don't need param lengths since text */
    NULL,    /* default to all text params */
    1);      /* ask for binary results */


  if (PQresultStatus(res) != PGRES_TUPLES_OK)
    {
    if(debug) syslog(LOG_NOTICE, "SELECT failed: %s", PQerrorMessage(variables->conn));
    fprintf(stderr, "SELECT failed: %s", PQerrorMessage(variables->conn));
    PQclear(res);
    return SMFIS_TEMPFAIL;
    }

  PQntuples_var = PQntuples(res);
  PQclear(res);
  
  if(debug) syslog(LOG_NOTICE, "SELECT Success. PQntuples(res)=%d", PQntuples_var);
  
  if (PQntuples_var > 0) 
    {
    if(debug) syslog(LOG_NOTICE, "Continue checking on next stage.");
    return SMFIS_CONTINUE;
    }
  else
    {
    if(debug) syslog(LOG_NOTICE, "local address check failed: <%s> not exist", variables->clean_address);
    sprintf(variables->temp, "%s: <%s>", bounce, variables->clean_address);
    smfi_setreply(ctx, "550", "5.1.1", variables->temp);
    if(debug) syslog(LOG_NOTICE, "Message rejected.");
    return SMFIS_REJECT;
    }
}

sfsistat
mlfi_envfrom(ctx, envfrom)
	SMFICTX *ctx;
	char **envfrom;
{
if(ignore_mail_from_cheking)
  return SMFIS_CONTINUE;
else
  return check_local_user_existing(ctx, envfrom, "MAIL FROM", "Fuck off, spoofer");
}

sfsistat
mlfi_envto(ctx, envto)
	SMFICTX *ctx;
	char **envto;
{
if(ignore_rcpt_to_cheking)
  return SMFIS_CONTINUE;
else
  return check_local_user_existing(ctx, envto, "RCPT TO", "I don't know about this e-mail address");
}


sfsistat
mlfi_close(ctx)
	SMFICTX *ctx;
{
	return mlfi_cleanup(ctx, true);
}

sfsistat
mlfi_cleanup(ctx, ok)
	SMFICTX *ctx;
	bool ok;
{

  disconnect_from_db(ctx);
  free_cortege_memory(ctx);
  if(smfi_getpriv(ctx)) free(smfi_getpriv(ctx));
  smfi_setpriv(ctx, NULL);     

  if(debug) syslog(LOG_NOTICE, "Work was complete.");
  return SMFIS_CONTINUE;
}

struct smfiDesc smfilter =
{
	"Local RCPT Checker (4 sendmail-BluePgSQL mod)",	/* filter name */
	SMFI_VERSION,	/* version code -- do not change */
	SMFIF_ADDHDRS,	/* flags */
	NULL,		/* connection info filter */
	NULL,		/* SMTP HELO command filter */
	mlfi_envfrom,	/* envelope sender filter */
	mlfi_envto,	/* envelope recipient filter */
	NULL,		/* header filter */
	NULL,		/* end of header */
	NULL,		/* body block filter */
	NULL,		/* end of message */
	NULL,		/* message aborted */
	mlfi_close	/* connection cleanup */
};

void term_handler(int i)
  {
  free(conninfo);
  if(debug) syslog(LOG_NOTICE, "Bye-bye...");
  exit(0);
  }


int
main(argc, argv)
	int argc;
	char *argv[];
{
//    
int c;

bool conntosendmail = false, conntopgsql = false;
debug = false;

// :        
openlog(NULL, LOG_PID, LOG_MAIL|LOG_USER);

conninfo = malloc(2048);
if(!conninfo) 
  {
  if(debug) syslog(LOG_NOTICE, "Can't alloc memory for \'conninfo\' variable");
  exit(1);
  }
// c -  
// s -  
// f -       
// h - 
// d -  
// i -  
	const char *args = "f:s:c:i:hd";

	/* Process command line options */
	while ((c = getopt(argc, argv, args)) != -1)
	{
		switch (c)
		{
		  case 'c':
			if (optarg == NULL || *optarg == '\0')
			{
				(void) fprintf(stderr, "Illegal conn: %s\n",
					       optarg);
				exit(EX_USAGE);
			}
                        conntosendmail = true;
			char *ofile;
			if (!strncmp(optarg, "unix:", 5))
        		    ofile = optarg + 5;
	    		else
		    	    if (!strncmp(optarg, "local:", 6)) ofile = optarg + 6;
			if (ofile) unlink(ofile);
			(void) smfi_setconn(optarg);
			break;
		  case 'f':
			if (optarg == NULL || *optarg == '\0')
			{
				(void) fprintf(stderr, "Illegal file name: %s\n",
					       optarg);
				exit(EX_USAGE);
			}
			FILE *conffile = fopen(optarg, "r");
			if(!conffile) exit(EX_USAGE);
			fgets(conninfo, 2047, conffile);
			fclose(conffile);
			conntopgsql = true;
			break;
		  case 'h':
			printf ("Usage: lrc-milter -c {unix|local|inet}:path {-s \'string\'|-f path} [-d] [-h] [-i {s|r}]\n");
			printf ("Options:\n");
			printf ("   -c {unix|local|inet}:path - path to socket file or network address to connect to sendmail\n");
			printf ("   -s \'string\'               - connection string to PgSQL\n");
			printf ("                               Example: \'dbname=database host=localhost port=5432 user=username password=secretword\'\n");
			printf ("   -f path                   - path to file with connection string as shown above\n");
			printf ("   -i                        - No check mode: \'s\' - MAIL FROM or \'r\' - RCPT TO, but no both\n");
			printf ("   -d                        - debug mode\n");
			printf ("   -h                        - this help\n");
			exit(EX_USAGE);
			break;
		  case 'd':
			debug = true;
			break;
		  case 'i':
			if(strchr(optarg, 's')) ignore_mail_from_cheking = true;
			if(strchr(optarg, 'r') && !ignore_mail_from_cheking) ignore_rcpt_to_cheking = true;
			break;
		  case 's':
			if (optarg == NULL || *optarg == '\0')
			  {
			  (void) fprintf(stderr, "Illegal connection string for PgSQL: %s\n", optarg);
			  exit(EX_USAGE);
			  }
                        conntopgsql = true;
			strcpy((char *)conninfo, optarg);
			break;
		}
	}

if (!conntosendmail || !conntopgsql)
  {
  printf ("Usage: lrc-milter -c {unix|local|inet}:socket_path {-s \'connection string to PgSQL\'|-f path_to_file_with_connection_string} [-d] [-h]\n");
  exit(EX_USAGE);
  }

//  
struct sigaction sa;
sigset_t newset;
sigemptyset(&newset);
sigaddset(&newset, SIGHUP);
sigprocmask(SIG_BLOCK, &newset, 0);
sa.sa_handler = term_handler;
sigaction(SIGTERM, &sa, 0);

//
    if (daemon(0, 0)) {
//        fprintf(stderr, "daemonize failed: %s\n", strerror(errno));
        }
//     	
	smfi_settimeout(350);

//  
	if (smfi_register(smfilter) == MI_FAILURE)
	{
		fprintf(stderr, "smfi_register failed\n");
		exit(EX_UNAVAILABLE);
	}
   return smfi_main();
}

/* eof */

//$Revision: 0.23 $, Last updated $Date: 2007/08/06 12:29:00 $

